#!/usr/bin/env python3
# Version 1.1.1
import sys
import getopt
import math
from numpy import loadtxt
import matplotlib.pyplot as plt

float_mode = 0
gauss_mode = 1
print_data = 1
write_data = 0 #write to file (e.g. for gnuplot use)
show_plot = 1
save_plot = 0
figsize = (12, 7.2) #figure dimensions in inches
logmode = 0 # 0:linear, 1:logarithmic

indatafile = 'covid19-de.txt'
txadjust=-1.5

tmax = 500
n_break = 10

sigma_t = 3 #overall time uncertainty in days
nsigma = 5 #cut-off radius of smoothing kernel in sigmas

#npop = 1.e6
npop = 82.e6

R_0 = 2.1 #basic reproductive number

t_exposed = 4#time from infection to infectious phase
t_infectious = 6#time period of virus transfer
t_sick = 7#time period being acutely sick and isolated
t_cure = 8#time period for cure after acute phase
t_conf = 9
p_death = 0.012

#re-scaling due to anti-spread measures
pscale1 = 0.30
pscale2 = 0.10
pscale3 = 0.05
pscale4 = 0.01
pscale5 = 0.01
pscale6 = 0.50
#times for anti-spread measures
t_pscale1 = 3800
t_pscale2 = t_pscale1+14
t_pscale3 = t_pscale2+14
t_pscale4 = t_pscale3+14
t_pscale5 = t_pscale4+14
t_pscale6 = t_pscale5+280

ts_0 = t_exposed
ts_1 = ts_0+t_infectious
ts_2 = ts_1+t_sick
ts_3 = ts_2+t_cure
ts_4 = ts_3+t_conf

def cmdargs(argv):
    global float_mode, gauss_mode, print_data, write_data, show_plot, save_plot, logmode, parfile, use_parfile
    helppage = '\npandana.py [options]\n\n'
    helppage += 'Options:\n'
    helppage += '-h, --help\n'
    helppage += '-i, --ifile        optional parameter file\n'
    helppage += '-f, --floatmode    default = 0\n'
    helppage += '-g, --gaussmode    default = 1\n'
    helppage += '-d, --printdata    default = 0\n'
    helppage += '-w, --writedata    default = 0\n'
    helppage += '-p, --showplot     default = 1\n'
    helppage += '-s, --saveplot     default = 0\n'
    helppage += '-l, --logmode      default = 0\n'
    try:
        opts, args = getopt.getopt(argv,"hi:f:g:d:w:p:s:l:",
                                   ["ifile=","floatmode=","gaussmode=","writedata=",
                                    "showplot=","saveplot=","logmode="])
    except getopt.GetoptError:
        print(helppage)
        sys.exit(2)
    use_parfile = False
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print(helppage)
            sys.exit()
        elif opt in ("-i", "--ifile"):
            parfile = arg
            use_parfile = True
        elif opt in ("-f", "--floatmode"):
            float_mode = int(arg)
        elif opt in ("-g", "--gaussmode"):
            gauss_mode = int(arg)
        elif opt in ("-d", "--printdata"):
            print_data = int(arg)
        elif opt in ("-w", "--writedata"):
            write_data = int(arg)
        elif opt in ("-p", "--showplot"):
            show_plot = int(arg)
        elif opt in ("-s", "--saveplot"):
            save_plot = int(arg)
        elif opt in ("-l", "--logmode"):
            logmode = int(arg)

cmdargs(sys.argv[1:])

if use_parfile:
    try:
        params = loadtxt(parfile)
    except:
        print('File does not exist: ', parfile)
        sys.exit()
    #print(params)
    #sys.exit()
    #population
    npop = params[0]
    #infection params
    R_0 = params[1]
    t_exposed = int(params[2])
    t_infectious = int(params[3])
    t_sick = int(params[4])
    t_cure = int(params[5])
    t_conf = int(params[6])
    p_death =  params[7]
    #re-scaling due to anti-spread measures
    pscale1 = params[8]
    pscale2 = params[9]
    pscale3 = params[10]
    pscale4 = params[11]
    pscale5 = params[12]
    pscale6 = params[13]
    #times for anti-spread measures
    t_pscale1 = int(params[14])
    t_pscale2 = int(params[15])
    t_pscale3 = int(params[16])
    t_pscale4 = int(params[17])
    t_pscale5 = int(params[18])
    t_pscale6 = int(params[19])
    #control params
    tmax = int(params[20])
    n_break = int(params[21])
    txadjust = params[22]
    #smoothing params
    sigma_t = int(params[23])
    nsigma = int(params[24])

#### initial conditions
p_infect_0 = R_0**(1/t_infectious)#daily reproduction rate
n_exposed = 1#start value
n_infectious = 0
n_infected = n_exposed
n_sick = 0
n_postacute = 0
n_survived = 0
n_dead = 0
n_cured = 0
n_conf = 0
n_ccum = 0

nprev_infected = 0
nprev_ccum = 0

ad_infected = []
ad_cured = []

for t in range(ts_4):
    ad_infected.append(0)

for t in range(ts_4):
    ad_cured.append(0)

ad_infected[0] = n_exposed

tmin = ts_4

rpsf = nsigma*sigma_t
a_psf = []
sf = 0
for i in range(2*rpsf):
    x = i-rpsf
    f = math.exp(-0.5*(x/sigma_t)**2)
    sf += f
    a_psf.append(f)
    
for i in range(2*rpsf):
    a_psf[i] /= sf
    #print('i, f = ', i, a_psf[i])

a_infected = []
a_conf = []
a_ccum = []
a_cured = []
a_dead = []
a_affected = []
a_change = []
a_R_eff = []
for t in range(tmax+1):
    a_infected.append(0)
    a_conf.append(0)
    a_ccum.append(0)
    a_cured.append(0)
    a_dead.append(0)
    a_affected.append(0)
    a_change.append(0)
    a_R_eff.append(0)

dfpart = 0
dfcure = 0

if gauss_mode > 0.5:
    tbreak = rpsf
else:
    tbreak = 0

tt = 0

#### Main loop
for t in range(tmax+1):
    if t > t_pscale6:
        pscale_infect = pscale6*p_infect_0
    elif t > t_pscale5:
        pscale_infect = pscale5*p_infect_0
    elif t > t_pscale4:
        pscale_infect = pscale4*p_infect_0
    elif t > t_pscale3:
        pscale_infect = pscale3*p_infect_0
    elif t > t_pscale2:
        pscale_infect = pscale2*p_infect_0
    elif t > t_pscale1:
        pscale_infect = pscale1*p_infect_0
    else:
        pscale_infect = p_infect_0

    f_exposed = n_exposed/npop
    f_infectious = n_infectious/npop
    f_sick = n_sick/npop
    f_infected = n_infected/npop
    f_dead = n_dead/npop
    f_survived = n_survived/npop
    f_postacute = n_postacute/npop
    f_cured = n_cured/npop
    f_conf = n_conf/npop
    f_ccum = n_ccum/npop
    
    #f_infected = f_exposed+f_infectious+f_sick
    f_affected = f_infected+f_postacute
    f_nonaff = 1-f_affected

    #f_conf = f_sick
    #f_conf = f_infectious+f_sick

    if t>tmin and n_infectious < n_break:
        p_infect = 0
    else:
        #p_infect = npop * pscale_infect * f_infectious * f_nonaff
        p_infect = npop * (1-(1-f_infectious*f_nonaff)**pscale_infect)

    if float_mode > 0.5:
        d_infect = p_infect
    else:
        dfpart += p_infect
        d_infect = int(dfpart)
        dfpart -= d_infect
    #print('d_infect = ', d_infect)
    #now... [0...ts_0-1...ts_1-1...ts_2-1] ...days ago
    d_infectious = ad_infected[ts_0-1]
    d_sick = ad_infected[ts_1-1]
    d_postacute = ad_infected[ts_2-1]
    d_conf = ad_infected[ts_0+t_conf-1]
    d_clean = ad_infected[ts_3-1]
    #d_cure = ad_cured[t_sick-1]
    
    ad_infected = [d_infect]+ad_infected[:-1]
    ad_cured = [n_survived]+ad_cured[:-1]
    daily_change = (n_infected-nprev_infected)/max(nprev_infected, 1)
    daily_ccum = (n_ccum-nprev_ccum)/max(nprev_ccum, 1)
    
    R_eff = pscale_infect**t_infectious#experimental

    a_infected[t] = f_infected
    a_conf[t] = f_conf
    a_ccum[t] = f_ccum
    a_cured[t] = f_cured
    a_dead[t] = f_dead
    a_affected[t] = f_affected
    #a_change[t] = daily_change
    a_change[t] = daily_ccum
    a_R_eff[t] = R_eff

    if tt > tbreak: break

    if t > tmin and max(n_conf, n_infected) <= 0.5:
        tt += 1
        print('t, tt, tbreak = ', t, tt, tbreak)

    nprev_infected = n_infected
    nprev_ccum = n_ccum
    n_exposed += d_infect#new infections
    n_exposed -= d_infectious
    n_infectious += d_infectious
    n_infectious -= d_sick#sick -> into hospital/quarantine
    n_sick += d_sick#new sick
    n_sick -= d_postacute#cured or dead
    n_postacute += d_postacute
    n_dead = int(p_death*n_postacute+0.5)
    n_survived = n_postacute-n_dead
    #n_cured += d_cure
    n_cured = ad_cured[t_cure-1]
    n_infected = n_exposed+n_infectious+n_sick
    n_ccum += d_conf
    n_conf += d_conf
    n_conf -= d_clean
    tend = t

#### Gaussian smoothing
if gauss_mode > 0.5:
    b_infected = []
    b_conf = []
    b_ccum = []
    b_cured = []
    b_dead = []
    b_affected = []
    b_change = []
    b_R_eff = []
    for t in range(tend+2*rpsf):
        b_infected.append(0)
        b_conf.append(0)
        b_ccum.append(0)
        b_cured.append(0)
        b_dead.append(0)
        b_affected.append(0)
        b_change.append(0)
        b_R_eff.append(0)
    
    for t in range(tend+1):
        for i in range(2*rpsf):
            f = a_psf[i]
            ii = t+i
            b_infected[ii] += f*a_infected[t]
            b_conf[ii] += f*a_conf[t]
            b_ccum[ii] += f*a_ccum[t]
            b_cured[ii] += f*a_cured[t]
            b_dead[ii] += f*a_dead[t]
            b_affected[ii] += f*a_affected[t]
            b_change[ii] += f*a_change[t]
            b_R_eff[ii] += f*a_R_eff[t] 

    for t in range(tend+1):
        i = t+rpsf
        a_infected[t] = b_infected[i]
        a_conf[t] = b_conf[i]
        a_ccum[t] = b_ccum[i]
        a_cured[t] = b_cured[i]
        a_dead[t] = b_dead[i]
        a_affected[t] = b_affected[i]
        a_change[t] = b_change[i]
        a_R_eff[t] = b_R_eff[i]

if write_data > 0.5:
    faux = open('pandana-aux.gp', 'w')
    faux.write('npop={:f}\n'.format(npop))
    faux.close()
    fout = open('pandana.out', 'w')

af_t = []
af_infected = []
af_conf = []
af_ccum = []
af_cured = []
af_dead = []
af_affected = []
af_change = []
af_R_eff = []

fmax = 0
ln2 = math.log(2)
for t in range(tend+3-tbreak):
    f_infected = a_infected[t]
    f_conf = a_conf[t]
    f_ccum = a_ccum[t]
    f_cured = a_cured[t]
    f_dead = a_dead[t]
    f_affected = a_affected[t]
    daily_change = a_change[t]
    R_eff = a_R_eff[t]
    if abs(daily_change) > 0.0007 and f_ccum*npop > 1:
        doubling_time = ln2/daily_change
        s_doubling_time = '{:6.2f}'.format(doubling_time)
    else:
        doubling_time = 0
        s_doubling_time = ' --- '
    fmax = max(fmax, f_ccum*npop)
    af_t.append(t)
    af_infected.append(f_infected*npop)
    af_conf.append(f_conf*npop)
    af_ccum.append(f_ccum*npop)
    af_cured.append(f_cured*npop)
    af_dead.append(f_dead*npop)
    af_affected.append(f_affected*npop)
    af_change.append(daily_change)
    af_R_eff.append(R_eff)
    if print_data > 0.5:
        print('{:4d} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:s}'.format(t, f_infected, f_conf, f_cured, f_dead, f_affected, f_ccum, daily_change, s_doubling_time))
    if write_data > 0.5:
        fout.write('{:4d} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:15.12f} {:s}\n'.format(t, f_infected, f_conf, f_cured, f_dead, f_affected, f_ccum, daily_change, s_doubling_time))

if write_data > 0.5:
    fout.close()

#### Plotting section
indata = loadtxt(indatafile)
xdata = indata[:, 0]+txadjust
ydata = indata[:, 1]

plt.figure(figsize=figsize)

dtop = 0.5
y10 = 10**int(math.log10(fmax))
k10 = fmax/y10
if logmode > 0.5:
    if k10 >= 4:
        d10 = 2*y10
    else:
        d10 = y10
    ymin = 1
    ymax = d10*math.ceil(fmax/d10+dtop)
    plt.yscale("log")
else:
    if k10 >= 5:
        d10 = 0.5*y10
    elif k10 >= 2:
        d10 = 0.2*y10
    else:
        d10 = 0.1*y10
    ymin = 0
    ymax = d10*math.ceil(fmax/d10+dtop)
    plt.yscale("linear")
#print('fmax, y10, k10, d10, ymax = ', fmax, y10, k10, d10, ymax)

plt.xlim(left=0)
plt.xlim(right=tend+3-tbreak)
plt.ylim(top=ymax)
plt.ylim(bottom=ymin)
plt.plot(af_t, af_infected, 'b-', label='infected')
plt.plot(af_t, af_cured, 'g-', label='cured')
plt.plot(af_t, af_dead, 'k-', label='dead')
plt.plot(af_t, af_conf, 'r-', label='confirmed')
plt.plot(af_t, af_affected, 'c-', label='cumulative')
plt.plot(af_t, af_ccum, 'm-', label='cum. conf.')
plt.plot(xdata, ydata, 'k+', label='COVID-19 in Germany')
plt.xlabel('days')
plt.ylabel('cases')
plt.legend()
plt.tight_layout()
if save_plot > 0.5:
    plt.savefig('pandana.png')
if show_plot > 0.5:
    plt.show()
